#!/usr/bin/env python
"""Programming-by-contract for Python, based on Eiffel's DBC.

Programming by contract documents class and modules with invariants,
expressions that must be true during the lifetime of a module or
instance; and documents functions and methods with pre- and post-
conditions that must be true during entry and return.

Copyright (c) 2003, Terence Way
This module is free software, and you may redistribute it and/or modify
it under the same terms as Python itself, so long as this copyright message
and disclaimer are retained in their original form.

IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF
THIS CODE, EVEN IF THE AUTHOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE.  THE CODE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS,
AND THERE IS NO OBLIGATION WHATSOEVER TO PROVIDE MAINTENANCE,
SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
"""

__author__ = "Terence Way"
__email__ = "terry@wayforward.net"
__version__ = "1.0 beta: May 21, 2003"
MODULE = 'contract'

import inspect
import new
import re
import tokenize

from cStringIO import StringIO
from types import *

# Programming-by-contract extends the language to include invariant
# expressions for classes, and pre-condition and post-condition
# expressions for methods.  These are very similar to assertions,
# expressions which must be true or the program is stopped.
#
# Class invariants are expressions which must be true at the end of a
# class's constructor, and at the beginning and end of each public
# method.
#
# Pre-conditions must be true at the start of each method.  Post-
# conditions must be true at the end of each method.  Post-condition
# logic has access to an 'old' variable which records variable values
# before they were changed by the method.
#
# This implementation tries to fix what I consider a defect in
# Eiffel's DBC -- while it is easy to specify what does change after
# invoking a method, it is impossible to specify what doesn't change.
# This makes it hard to reason about methods, and limits portability
# to more formal techniques such as Z.
#
# A simple fix is to declare what parts change.  Simply modify the
# line 'post:' to 'post [a]:' to declare that the variable a is the
# only thing changed by the sort function.  This is modelled after Z
# schemas [2].
#
# Simply saying 'post:' means anything can be changed: this is the
# original semantics of the Eiffel ENSURES statement.  Saying 'post
# []:' means nothing is changed.
#
# Python Specifications
# ---------------------
# The docstring of any module or class can have invariants documented
# by having the keyword 'inv' followed by a colon (:) at the start of
# a line.  Whitespace is ignored.  This is either followed by a single
# expression on the same line, or a series of expressions on following
# lines indented past the 'post' key- word.  The normal Python rules
# about implicit and explicit line continuations are followed here.
#
# Module invariants must be true at module load time, and at the entry
# and return of every public function within the module.
#
# Class invariants must be true after the __init__ function returns,
# and at the entry and return of every public method of the class.
#
# The docstring of any function or method can have pre-conditions
# documented with the keyword 'pre' following the same rules above,
# and can have post-conditions documented with the keyword 'post'
# optionally followed by a list of variables.  The variables are in
# the same scope as the body of the function or method.
#
# Expressions have access to some additional convenience values.  To
# make evaluating sequences easier, there are two functions: forall(a)
# and exists(a).  To make implication easier, there is a
# implies(x,a,b=True) function which mirrors C's conditional
# expression (x?a:c).
#
# The expressions in the post-conditions have access to two additional
# variables, __old__ which is filled with variables declared in the
# 'post' statement bound to shallow copies before the function or
# method is called; and __return__ which is bound to the return value
# of the function or method.
#
# Implementation
# --------------
# This module is divided into four parts:
#
# 1 -- Find docstrings and parse them for contract expressions...
#
#      This is accomplished by:
#      a. recursive enumerating elements in modules, classes using code
#         lifted from the 'inspect' Python standard module
#         see: checkmod, _check
#      b. scanning the docstrings of public functions with a regular
#         expression looking for lines that matches 'pre:' 'post:' or
#         'inv:' at the start
#         see: parse_docstring
#      c. Using the 'tokenize' Python tokenizer to build expressions
#         see: _read_block, _read_decl, _read_token
#
# 2 -- Construct functions that do contract checking...
#      This is done by just constructing big strings that are function
#      definitions.  Each function or method gets associated with four
#      'helper' functions... to check pre-conditions, to save old
#      values, to check post-conditions, and of course the saved
#      original code.  These are stored as function attributes.
#         see: _install_wrapper, _define_checker
#
# 3 -- Run-time support for call checking
#         see: call_public_funcion, call_private_function,
#              call_public_method, call_private_method,
#              call_constructor, call_destructor
#
# 4 -- Declarations for use within contract expressions
#      includes 'forall' 'exists' 'implies' and others


# change the keywords here, if necessary, and in the _re_start definition below
INVARIANTS = 'inv'
CONDITIONS = ['pre', 'post']
OLD = '__old__'
RETURN = '__return__'
PREFIX = '__assert_'

_ORIG = 'orig'
_SAVE = 'save'
_CHK = 'chk'

# look for word+ [expr] space* ':' at the start of a line
#
_re_start = re.compile(r'^\s*(pre|post|inv)\s*(\[|:)', re.MULTILINE)
_RE_KEYWORD = 1

# the tokenizer only seems to return OP(50) for any operators like : [ ] . ,
# we need to do a further lookup to get the operator value, but only for the
# tokens we care about
#
_OPS = {':': tokenize.COLON,
        '[': tokenize.LSQB,
        ']': tokenize.RSQB,
        '.': tokenize.DOT,
        ',': tokenize.COMMA}

_PROC_TYPES = (MethodType, FunctionType)
_TYPE_TYPES = (TypeType, ClassType)

###############################################################################
# Part 1 -- Find docstrings and parse them for contract expressions...
#

def checkmod(module):
    """Add invariant, pre- and post-condition checking to a module.

    pre:
        isinstance(module, ModuleType)
    """
    if __debug__:
        # get members *before* we start adding stuff to this module
        members = _get_members(module)
        invs = parse_docstring(module.__doc__, [INVARIANTS])[0]
        name = PREFIX + INVARIANTS
        func = getattr(module, name, None)
        # should we override
        if not func or func.__name__.startswith(PREFIX):
            if invs[2]:
                func = _define_checker(name, '', invs, module)
            else:
                func = __assert_inv
                module.__assert_inv = func
        _check_members(members, [module])
        # check module invariants now
        func()

def _check_type(code, path):
    """Modify a module or class to add invariant checking.

    pre:
        isinstance(code[0], StringTypes)
        type(code[1]) == code[2]
        isinstance(code[2], _CODE_TYPES)
        isinstance(path[0], ModuleType)
        forall(path[1:], lambda x: isinstance(x, _TYPE_TYPES))
    """
    name, obj = code

    # get members *before* we start adding stuff to this class
    members = _get_members(obj)
    path = path + [obj]
    invs = parse_docstring(obj.__doc__, [INVARIANTS])[0]
    if invs[2]:
        func = _define_checker(_mkname(path, INVARIANTS), 'self', invs,
                               path[0])
        setattr(obj, PREFIX + INVARIANTS, func)
        delattr(path[0], func.func_name)
    _check_members(members, path)

def _check_proc(code, path):
    """Modify a module or class to add invariant checking.
    """
    name, obj = code
    assert isinstance(obj, _PROC_TYPES)
    _install_wrapper(code, parse_docstring(obj.__doc__, CONDITIONS),
                     path, is_public = _ispublic(name))

def _get_members(obj):
    """Returns two lists (procs, types) where each list contains (name,
    value) tuples.

    For classes, only attributes defined by the specific class are returned,
    i.e. not inherited attributes.  Attributes created by this module
    (prefixed by '__assert_') are skipped as well.

    >>> import contract
    >>> hasattr(contract, '_re_start')
    1
    >>> '_re_start' in [x[0] for x in _get_members(contract)[0]]
    0
    >>> '_get_members' in [x[0] for x in _get_members(contract)[0]]
    1
    >>> 'checkmod' in [x[0] for x in _get_members(contract)[0]]
    1
    >>> class base:
    ...     def foo(self): pass
    >>> class derived(base):
    ...     def bar(self): pass
    >>> hasattr(derived, 'foo') # hasattr can get inherited attributes
    1
    >>> 'foo' in [x[0] for x in _get_members(derived)[0]] # but we don't
    0
    """
    procs = []
    types = []
    for key in obj.__dict__:
        if not key.startswith(PREFIX):
            value = getattr(obj, key)
            if isinstance(value, _PROC_TYPES):
                procs.append((key, value))
            elif isinstance(value, _TYPE_TYPES):
                types.append((key, value))
    return (procs, types)

def _check_members((procs, types), path):
    for p in procs:
        _check_proc(p, path)
    for t in types:
        _check_type(t, path)

def _ispublic(name):
    """Checks if a name is public (starts and ends with '__' or doesn't
    start with a _ at all).

    >>> _ispublic('__init__')
    1
    >>> _ispublic('foo')
    1
    >>> _ispublic('_ispublic')
    0
    """
    return not name.startswith('_') or \
        (name.startswith('__') and name.endswith('__'))

def parse_docstring(docstring, keywords):
    """Parse a docstring, looking for design-by-contract expressions.

    Returns a list of tuples: the list is the same length as keywords, and
    matches each keyword.  The tuple is (keyword, [decls], [exprs]), namely
    the keyword, a list of string declarations, and a list of string
    expressions.

    Example:
    >>> from pprint import pprint
    >>> pprint( parse_docstring(parse_docstring.__doc__, ['post', 'pre']) )
    [('post', [], ['[ x [ 0 ] for x in __return__ ] == keywords']),
     ('pre',
      [],
      ['docstring is None or isinstance ( docstring , StringTypes )',
       'forall ( keywords , lambda x : isinstance ( x , StringTypes ) )'])]

    pre:
        docstring is None or isinstance(docstring, StringTypes)
        forall(keywords, lambda x: isinstance(x, StringTypes))

    post[]:
        [x[0] for x in __return__] == keywords
    """
    result = [(x, [], []) for x in keywords]

    if docstring is None:
        return result

    # step 1: scan through docstring looking for keyword
    input = StringIO(docstring)

    offset = 0
    assert input.tell() == 0

    line = input.readline()
    while line != '':
        a = _re_start.split(line)

        if len(a) > _RE_KEYWORD and a[_RE_KEYWORD] in keywords:
            # step 2: found a keyword, now rewind and scan looking
            # for either an inline expression or a series of sub-
            # indented expressions
            past = input.tell()
            input.seek(offset)

            l = _read_block(input)
            # returns (keyword, decls, exprs)

            # Find the right result index based on keyword
            r = result[keywords.index(l[0])]
            r[1].extend(l[1])
            r[2].extend(l[2])

            # Unfortunately, the Python tokenizer eats lines
            # looking for the DEDENT... so we seek back past
            # the line containing the keyword
            offset = past
            input.seek(offset)
        else:
            offset = input.tell()
        line = input.readline()
    return result

# our alphabet:
#    INDENT(5), DEDENT(6), NAME(1), NEWLINE(4), OP(50), COMMENT(52), NL(53)
#
# block: inv | pre | post
# inv:   INDENT 'inv' ':' suite
# pre:   INDENT 'pre' ':' suite
# post:  INDENT 'post' [ '[' decls ']' ] ':' suite
# decls: /* empty */ | decl ( ',' decl )*
# decl:  NAME ( '.' NAME )*
#
# suite:   NEWLINE INDENT stuff DEDENT
#        | stuff NEWLINE
#
def _read_block(input):
    r"""Read an indented block of expressions

    pre:
        input.readline  # must have readline function

    Examples:
    >>> _read_block(StringIO('\tfoo:\n'))
    0
    >>> _read_block(StringIO('\tpost[]: True\n'))
    ('post', [], ['True'])
    >>> _read_block(StringIO('\tpre: 5 + 6 > 10\n'))
    ('pre', [], ['5 + 6 > 10'])
    >>> _read_block(StringIO('\tpost:\n\t\t5 + 6 < 12\n\t\t2 + 2 == 4'))
    ('post', [], ['5 + 6 < 12', '2 + 2 == 4'])
    >>> _read_block(StringIO('\tpost[foo.bar]: # changes\n' \
    ...                      '\t\tlen(foo.bar) > 0\n'))
    ('post', [['foo', 'bar']], ['len ( foo . bar ) > 0'])
    """
    g = tokenize.generate_tokens(input.readline)
    t = g.next()
    if t[0] == tokenize.INDENT:
        t = g.next()
    if t[0] != tokenize.NAME:
        return False
    keyword = t[1]
    t = _read_token(g)

    # handle declarations in [name1, name2.field] format
    if t[0] == tokenize.LSQB:
        decls = _read_decl(g)
        if decls is None:
            return False
        t = _read_token(g)
    else:
        decls = []

    if t[0] != tokenize.COLON:
        return False
    t = _read_token(g)

    # two choices, we've read a newline, result is a list of all lines
    # indented under us (tokenize will pick this up).  Otherwise we're
    # reading an expression until the newline
    #
    if t[0] == tokenize.NEWLINE:
        t = _read_token(g)
        if t[0] != tokenize.INDENT:
            return False
        e = tokenize.DEDENT
    else:
        e = tokenize.NEWLINE

    # now concatenate token strings until we reach the end
    result = []
    string = ''
    while t[0] != e and t[0] != tokenize.ENDMARKER:
        if t[0] == tokenize.NEWLINE:
            result.append(string.strip())
            string = ''
        else:
            string += t[1] + ' '
        t = _read_token(g)
    if string:
        result.append(string.strip())
    return (keyword, decls, result)

def _read_decl(g):
    """Returns a list of list of identifiers [['name', 'name'], ...] or
    None on error.
    """
    decl = []
    result = []
    t = _read_token(g)
    while t[0] != tokenize.RSQB:
        if t[0] == tokenize.NAME:
            decl.append(t[1])
        elif t[0] == tokenize.COMMA:
            result.append(decl)
            decl = []
        elif t[0] != tokenize.DOT:
            return None
        t = _read_token(g)
    if decl:
        result.append(decl)
    return result

def _read_token(g):
    """Python tokenizer, skips comments and implicit newlines.

    Also replaces select tokenize.OP with real token values.
    """
    t = g.next()
    while t[0] == tokenize.COMMENT or t[0] == tokenize.NL:
        t = g.next()
    if t[0] == tokenize.OP and t[1] in _OPS:
        return (_OPS[t[1]], t[1], t[2], t[3], t[4])
    else:
        return t

# ...part 1
###############################################################################

###############################################################################
# Part 2 -- Construct functions that do contract checking...
#
def _install_wrapper(code, contracts, path, is_public):
    """Creates and installs a function/method checker.

    pre:
        contracts[0][0] == 'pre' and contracts[1][0] == 'post'
        isinstance(path[0], ModuleType)
        forall(path[1:], lambda x: isinstance(x, _TYPE_TYPES))
    """
    name, obj = code

    if isinstance(obj, MethodType):
        func = obj.im_func
        invs = hasattr(path[-1], PREFIX + INVARIANTS)
    else:
        assert isinstance(obj, FunctionType)
        func = obj
        invs = hasattr(path[0], PREFIX + INVARIANTS)

    # we must create a checker if:
    #  1. there are any pre-conditions or any post-conditions OR
    #  2. this is public AND there are invariants
    if contracts[0][2] or contracts[1][2] or (is_public and invs):
        argspec = inspect.getargspec(func)
        args = _format_args(argspec)

        # argl: argument list suitable for appending at the end of other args
        if args:
            argl = ', ' + args
        else:
            argl = args

        output = StringIO()

        output.write('def %s(%s):\n' % (_mkname(path, name, _CHK), args))
        output.write('\timport %s\n' % MODULE)

        output.write('\treturn %s.call_' % MODULE)

        if isinstance(obj, FunctionType):
            if is_public:
                output.write('public_function(__assert_inv, ')
            else:
                output.write('private_function(')
        else:
            if not is_public:
                output.write('private_method(')
            elif name == '__init__':
                output.write('constructor(')
            elif name == '__del__':
                output.write('destructor(')
            else:
                output.write('public_method(')

            output.write('.'.join([c.__name__ for c in path[1:]]))
            output.write(', ')

        for c in path[1:]:
            output.write(c.__name__)
            output.write('.')
        output.write(name)
        output.write(argl)
        output.write(')\n')

        newfunc = _define(_mkname(path, name, _CHK), output.getvalue(),
                          path[0])

        # if there are default arguments
        if argspec[3]:
            # install the default arguments... don't try to put them into
            # our printed function definition, above, as they have already
            # been evaluated.
            newfunc = new.function(newfunc.func_code, newfunc.func_globals,
                                   newfunc.func_name, argspec[3])

        setattr(newfunc, PREFIX + _ORIG, getattr(func, PREFIX + _ORIG, func))

        # write preconditions checker
        if contracts[0][2]:
            pre = _define_checker(_mkname(path, name, contracts[0][0]), args,
                                  contracts[0], path[0])

            delattr(path[0], pre.func_name)
            setattr(newfunc, PREFIX + contracts[0][0], pre)

        # write __old__ saver
        if contracts[1][1]:
            saver = _define_saver(_mkname(path, name, _SAVE), OLD + argl,
                                  contracts[1][1], path[0])
            delattr(path[0], saver.func_name)
            setattr(newfunc, PREFIX + _SAVE, saver)

        # write postconditions checker
        if contracts[1][2]:
            post = _define_checker(_mkname(path, name, contracts[1][0]),
                                   OLD + ', ' + RETURN + argl, contracts[1],
                                   path[0])

            delattr(path[0], post.func_name)
            setattr(newfunc, PREFIX + contracts[1][0], post)

        newname = newfunc.func_name
        newfunc.__doc__ = obj.__doc__

        if isinstance(path[-1], _TYPE_TYPES) and isinstance(obj, FunctionType):
            # static method
            newfunc = staticmethod(newfunc)

        setattr(path[-1], name, newfunc)

        if name != newname or len(path) > 1:
            delattr(path[0], newname)
    # end _install_wrapper

def _define_checker(name, args, contract, module):
    """Define a function that does contract assertion checking.

    args is a string argument declaration (ex: 'a, b, c = 1, *va, **ka')
    contract is an element of the contracts list returned by parse_docstring
    module is the containing module (not parent class)

    Returns the newly-defined function.

    pre:
        isinstance(name, StringTypes)
        isinstance(args, StringTypes)
        contract[0] in ['pre', 'post', 'inv']
        len(contract[2]) > 0
    post:
        isinstance(__return__, FunctionType)
        __return__.__name__ == name
    """
    output = StringIO()
    output.write('def %s(%s):\n' % (name, args))
    output.write('\tfrom %s import forall, exists, implies\n' % MODULE)
    for c in contract[2]:
        output.write('\tassert ')
        output.write(c)
        output.write(', "%s"\n' % contract[0])
    return _define(name, output.getvalue(), module)

def _define_saver(name, args, decls, module):
    """Create a function that saves values into an __old__ variable.

    pre: decls
    post: isinstance(__return__, FunctionType)
    """
    output = StringIO()
    output.write('def %s(%s):\n' % (name, args))
    output.write('\timport %s, copy\n' % MODULE)

    _save_decls(output, '', _decltodict(decls))

    return _define(name, output.getvalue(), module)

def _define(name, text, module):
    #print text
    exec text in vars(module)
    return getattr(module, name)

def _format_args( (arguments, rest, keywords, default_values) ):
    """Formats an argument desc into a string suitable for both a function/
    method declaration or a function call.

    This does *not* handle default arguments.  Default arguments are already
    evaluated... use new.function() to create a function with pre-evaluated
    default arguments.

    Examples:
    >>> import inspect
    >>> def foo(a, (b, c), d = 1, e = 2, *va, **ka):
    ...     pass
    >>> a = inspect.getargspec(foo)
    >>> a
    (['a', ['b', 'c'], 'd', 'e'], 'va', 'ka', (1, 2))
    >>> _format_args(a)
    'a, (b, c), d, e, *va, **ka'

    pre:
        isinstance(arguments, list)
        rest is None or isinstance(rest, StringTypes)
        keywords is None or isinstance(keywords, StringTypes)
    post[]: True
    """
    if rest is not None: arguments = arguments + ['*' + rest]
    if keywords is not None: arguments = arguments + ['**' + keywords]
    return ', '.join([_format_arg(i) for i in arguments])

def _format_arg(a):
    """Format argument variables or tuples.
    """
    if not isinstance(a, list):
        return a
    elif len(a) == 1:
        # tuples of length 1 have to be output as (val,)
        return '(' + _format_arg(a[0]) + ',)'
    else:
        return '(' + ', '.join([_format_arg(i) for i in a]) + ')'

def _mkname(path, *va):
    """Define a name combining a path and arbitrary strings.

    pre:
       isinstance(path[0], ModuleType)
       forall(path[1:], lambda x: isinstance(x, _TYPE_TYPES))

    >>> import contract
    >>> _mkname([contract], 'test')
    '__assert_test'
    >>> class foo:
    ...     pass
    >>> _mkname([contract, foo], 'func', 'pre')
    '__assert_foo_func_pre'
    """
    return PREFIX + '_'.join([c.__name__ for c in path[1:]] + list(va))

def _save_decls(output, name, d):
    """Recursively output a dictionary into a set of 'old' assignments.

    The dictionary d is a tree of variable declarations.  So, for example,
    the declaration [self, self.buf, self.obj.a] would turn into the dict
    {'self': {'buf': {}, 'obj': {'a': {}}}}, and would get output as
        __old__.self = contract._holder()
        __old__.self.buf = copy.copy(self.buf)
        __old__.self.obj = contract._holder()
        __old__.self.obj.a = copy.copy(self.obj.a)
    """
    for k, v in d.items():
        n = name + k
        output.write('\t%s.%s = ' % (OLD, n))
        if v == {}:
            output.write('copy.copy(%s)\n' % n)
        else:
            output.write('%s._holder()\n' % MODULE)
            _save_decls(output, n + '.', v)

def _decltodict(l):
    """Converts a list of list of names into a hierarchy of dictionaries.

    Examples:
    >>> d = _decltodict([['self', 'buf'],
    ...                  ['self', 'item', 'a'],
    ...                  ['self', 'item', 'b']])
    >>> d == {'self': {'buf': {}, 'item': {'a': {}, 'b': {}}}}
    1
    """
    result = {}
    for i in l:
        d = result
        for n in i:
            # n in d ? d[n] : d[n] = {}
            d = d.setdefault(n, {})
    return result

#
# ...part 2
###############################################################################

###############################################################################
# Part 3... Run-time support for call checking
#
def call_public_function(inv, func, *va, **ka):
    """Check the invocation of a function or static method.
    """
    inv()
    try:
        return call_private_function(func, *va, **ka)
    finally:
        inv()

def call_private_function(func, *va, **ka):
    return _call([func], func, *va, **ka)

def call_public_method(cls, method, *va, **ka):
    mro = inspect.getmro(cls)
    check_invariants(mro, va[0])
    try:
        return _method_call(mro, method, *va, **ka)
    finally:
        check_invariants(mro, va[0])

def call_constructor(cls, method, *va, **ka):
    mro = inspect.getmro(cls)
    try:
        return _method_call(mro, method, *va, **ka)
    finally:
        check_invariants(mro, va[0])

def call_destructor(cls, method, *va, **ka):
    mro = inspect.getmro(cls)
    check_invariants(mro, va[0])
    return _method_call(mro, method, *va, **ka)
    
def call_private_method(cls, method, *va, **ka):
    return _method_call(inspect.getmro(cls), method, *va, **ka)

def check_invariants(mro, instance):
    for c in mro:
        try:
            p = c.__assert_inv
        except AttributeError:
            pass
        else:
            p(instance)

def _method_call(mro, method, *va, **ka):
    """Check the invocation of a method.

    mro -- list/tuple of class objects in method resolution order
    """
    # NO CONTRACTS... recursion
    assert isinstance(method, MethodType)

    func = method.im_func
    name = func.__assert_orig.__name__
    # list of all method functions with name
    a = [getattr(c, name).im_func for c in mro if _has_method(c, name)]

    return _call(a, func, *va, **ka)

def _call(a, func, *va, **ka):
    assert isinstance(func, FunctionType)

    # check pre-conditions, at least one must pass
    pre = [f.__assert_pre for f in a if hasattr(f, '__assert_pre')]
    if pre:
        for p in pre[:-1]:
            try:
                p(*va, **ka)
                break
            except AssertionError:
                pass
        else:
            pre[-1](*va, **ka)

    # save old values
    old = _holder()
    for f in a:
        try:
            p = f.__assert_save
        except AttributeError:
            pass
        else:
            p(old, *va, **ka)

    result = func.__assert_orig(*va, **ka)
    for f in a:
        # check post-conditions
        try:
            p = f.__assert_post
        except AttributeError:
            pass
        else:
            p(old, result, *va, **ka)

    return result

def _has_method(cls, name):
    """Test if a class has a named method.

    pre: isinstance(cls, _TYPE_TYPES)
    pre: isinstance(name, StringTypes)
    post: __return__ == (hasattr(cls, name) and \
                         isinstance(getattr(cls, name), MethodType))
    """
    return isinstance(getattr(cls, name, None), MethodType)

def __assert_inv():
    """Empty invariant assertions
    """
    pass

# ...part 3
###############################################################################

###############################################################################
# Part 4 -- Declarations for use within contract expressions...
#
def forall(a, fn = bool):
    """Checks that all elements in a sequence are true.

    Returns True(1) if all elements are true.  Return False(0) otherwise.

    >>> forall([True, True, True])
    1
    >>> forall( () )
    1
    >>> forall([True, True, False, True])
    0
    """
    for i in a:
        if not fn(i):
            return False
    return True

def exists(a, fn = bool):
    """Checks that at least one element in a sequence is true.

    Returns True(1) if at least one element is true.  Return False(0)
    otherwise.

    >>> exists([False, False, True])
    1
    >>> exists([])
    0
    >>> exists([False, 0, '', []])
    0
    """
    for i in a:
       if fn(i):
           return True
    return False

def implies(test, then_val, else_val = True):
    """Logical implication.

    implies(x, y) should be read 'x implies y' or 'if x then y'
    implies(x, a, b) should be read 'if x then a else b'

    >>> implies(False, False)
    1
    >>> implies(False, True)
    1
    >>> implies(True, False)
    0
    >>> implies(True, True)
    1
    """
    if test:
        return then_val
    else:
        return else_val

class _holder:
    """Placeholder for arbitrary 'old' values.
    """
    pass

#
# ...part 4
###############################################################################

__test__ = {
    '_ispublic': _ispublic, '_get_members': _get_members,
    '_decltodict': _decltodict, '_read_block': _read_block,
    '_format_args': _format_args, '_mkname': _mkname}

if __name__ == '__main__':
    import doctest, contract
    #contract.checkmod(contract)
    doctest.testmod(contract)
